//! Rocks DB implementation of the storage layer
use crate::{event_filter::EventFilter, Error, RocksDb, Storage};
use nostr_rs_types::types::Event;
use rocksdb::{BoundColumnFamily, DBIteratorWithThreadMode, DB};
use std::{collections::VecDeque, sync::Arc};

pub struct ResultSet<'a> {
    pub db: &'a RocksDb,
    pub filter: EventFilter,
    pub namespace: Option<Arc<BoundColumnFamily<'a>>>,
    pub secondary_index_lookup: Option<DBIteratorWithThreadMode<'a, DB>>,
    pub current_prefix: Vec<u8>,
    pub prefixes: VecDeque<Vec<u8>>,
    pub limit: Option<usize>,
    pub returned: usize,
}

impl<'a> ResultSet<'a> {
    /// Selects the next prefix available and starts an iterator using the
    /// secondary index. If no prefix is available from prefixes the functions
    /// return None, signalling upstream the are no more results
    fn select_next_prefix_using_secondary_index(&mut self) -> Option<()> {
        let prefix = self.prefixes.pop_front()?;
        self.secondary_index_lookup = Some(
            self.db
                .db
                .prefix_iterator_cf(self.namespace.as_ref()?, prefix.clone()),
        );
        self.current_prefix = prefix;
        Some(())
    }
}

impl<'a> Iterator for ResultSet<'a> {
    type Item = Result<Event, Error>;

    fn next(&mut self) -> Option<Self::Item> {
        if Some(self.returned) == self.limit {
            return None;
        }
        if self.secondary_index_lookup.is_none() {
            if self.namespace.is_some() {
                self.select_next_prefix_using_secondary_index()?;
            } else {
                // No secondary index is used to query, this means the query is
                // using the ID filter, so it is more efficient to use the
                // primary index to prefetch events that may satisfy the query
                loop {
                    let prefix = self.prefixes.pop_front()?;
                    if let Ok(Some(event)) = self.db.get_event(prefix) {
                        if self.filter.check_event(&event) {
                            self.returned += 1;
                            return Some(Ok(event));
                        }
                    }
                }
            }
        }

        loop {
            loop {
                let secondary_index = self.secondary_index_lookup.as_mut()?;
                let (key, value) = match secondary_index.next() {
                    Some(Ok((k, v))) => (k, v),
                    _ => {
                        // break this loop to select next available prefix
                        break;
                    }
                };
                if !key.starts_with(&self.current_prefix) {
                    break;
                }
                if let Ok(Some(event)) = self.db.get_event(value) {
                    if self.filter.check_event(&event) {
                        self.returned += 1;
                        return Some(Ok(event));
                    }
                }
            }
            // Select next prefix if available, or exists
            self.select_next_prefix_using_secondary_index()?;
        }
    }
}
